# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: signals.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rsignals.proto\x12\x13graphsignal.signals\"|\n\rUploadRequest\x12*\n\x06traces\x18\x01 \x03(\x0b\x32\x1a.graphsignal.signals.Trace\x12,\n\x07metrics\x18\x03 \x03(\x0b\x32\x1b.graphsignal.signals.Metric\x12\x11\n\tupload_ms\x18\x02 \x01(\x04\"\x10\n\x0eUploadResponse\"\x90\x01\n\x0bTraceRecord\x12\x0f\n\x07\x64\x61ta_id\x18\x01 \x01(\t\x12\x13\n\x0b\x65ndpoint_id\x18\x02 \x01(\t\x12)\n\x05trace\x18\x04 \x01(\x0b\x32\x1a.graphsignal.signals.Trace\x12\x1a\n\x12\x64\x61ta_retention_sec\x18\x05 \x01(\x04\x12\x14\n\x0ctime_skew_ms\x18\x06 \x01(\x12\"\xba\x08\n\x05Trace\x12\x10\n\x08trace_id\x18\x30 \x01(\t\x12\x10\n\x08start_us\x18\x05 \x01(\x04\x12\x0e\n\x06\x65nd_us\x18\x06 \x01(\x04\x12\x0e\n\x06labels\x18\x41 \x03(\t\x12>\n\rsampling_type\x18\x42 \x01(\x0e\x32\'.graphsignal.signals.Trace.SamplingType\x12\'\n\x04span\x18? \x01(\x0b\x32\x19.graphsignal.signals.Span\x12&\n\x04tags\x18- \x03(\x0b\x32\x18.graphsignal.signals.Tag\x12*\n\x06params\x18; \x03(\x0b\x32\x1a.graphsignal.signals.Param\x12\x32\n\nexceptions\x18/ \x03(\x0b\x32\x1e.graphsignal.signals.Exception\x12<\n\ralloc_summary\x18< \x03(\x0b\x32%.graphsignal.signals.MemoryAllocation\x12\x30\n\nop_profile\x18= \x03(\x0b\x32\x1c.graphsignal.signals.OpStats\x12\x38\n\x0ekernel_profile\x18> \x03(\x0b\x32 .graphsignal.signals.KernelStats\x12\x34\n\x0c\x64\x61ta_profile\x18\x32 \x03(\x0b\x32\x1e.graphsignal.signals.DataStats\x12\x35\n\x0c\x64\x61ta_samples\x18@ \x03(\x0b\x32\x1f.graphsignal.signals.DataSample\x12\x32\n\nmodel_info\x18& \x01(\x0b\x32\x1e.graphsignal.signals.ModelInfo\x12\x36\n\nframeworks\x18$ \x03(\x0b\x32\".graphsignal.signals.FrameworkInfo\x12\x36\n\x0c\x64\x65vice_usage\x18\x12 \x03(\x0b\x32 .graphsignal.signals.DeviceUsage\x12\x32\n\nnode_usage\x18\x19 \x01(\x0b\x32\x1e.graphsignal.signals.NodeUsage\x12\x38\n\rprocess_usage\x18\x1a \x01(\x0b\x32!.graphsignal.signals.ProcessUsage\x12\x34\n\x0btracer_info\x18# \x01(\x0b\x32\x1f.graphsignal.signals.TracerInfo\x12\x37\n\rtracer_errors\x18\x07 \x03(\x0b\x32 .graphsignal.signals.TracerError\"d\n\x0cSamplingType\x12\x16\n\x12UNDEFINED_SAMPLING\x10\x00\x12\x13\n\x0fRANDOM_SAMPLING\x10\x01\x12\x13\n\x0fPARENT_SAMPLING\x10\x02\x12\x12\n\x0e\x45RROR_SAMPLING\x10\x03\"!\n\x03Tag\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"$\n\x05Param\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"\xa2\x01\n\x04Span\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08start_ns\x18\x02 \x01(\x04\x12\x0e\n\x06\x65nd_ns\x18\x03 \x01(\x04\x12\x10\n\x08trace_id\x18\x07 \x01(\t\x12\x15\n\rroot_trace_id\x18\x08 \x01(\t\x12\x17\n\x0fparent_trace_id\x18\t \x01(\t\x12(\n\x05spans\x18\x05 \x03(\x0b\x32\x19.graphsignal.signals.Span\"C\n\tException\x12\x10\n\x08\x65xc_type\x18\x01 \x01(\t\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x13\n\x0bstack_trace\x18\x03 \x01(\t\"5\n\x06SemVer\x12\r\n\x05major\x18\x01 \x01(\x05\x12\r\n\x05minor\x18\x02 \x01(\x05\x12\r\n\x05patch\x18\x03 \x01(\x05\"\xc2\x02\n\x10MemoryAllocation\x12K\n\x0e\x61llocator_type\x18\x01 \x01(\x0e\x32\x33.graphsignal.signals.MemoryAllocation.AllocatorType\x12\x12\n\ndevice_idx\x18\x02 \x01(\x04\x12\x16\n\x0e\x61llocated_size\x18\x03 \x01(\x04\x12\x15\n\rreserved_size\x18\x04 \x01(\x04\x12\x12\n\nfreed_size\x18\x05 \x01(\x04\x12\x17\n\x0fnum_allocations\x18\x06 \x01(\x04\x12\x19\n\x11num_alloc_retries\x18\x07 \x01(\x04\x12\x10\n\x08num_ooms\x18\x08 \x01(\x04\"D\n\rAllocatorType\x12\x17\n\x13UNDEFINED_ALLOCATOR\x10\x00\x12\x1a\n\x16PYTORCH_CUDA_ALLOCATOR\x10\x01\"\xcd\x03\n\x07OpStats\x12\x34\n\x07op_type\x18\x01 \x01(\x0e\x32#.graphsignal.signals.OpStats.OpType\x12\x0f\n\x07op_name\x18\x02 \x01(\t\x12\r\n\x05\x63ount\x18\x03 \x01(\x04\x12\x14\n\x0chost_time_ns\x18\x04 \x01(\x04\x12\x19\n\x11self_host_time_ns\x18\x05 \x01(\x04\x12\x1e\n\x16self_host_time_percent\x18\x06 \x01(\x01\x12\x16\n\x0e\x64\x65vice_time_ns\x18\x07 \x01(\x04\x12\x1b\n\x13self_device_time_ns\x18\x08 \x01(\x04\x12\x13\n\x0bhost_memory\x18\t \x01(\x04\x12\x18\n\x10self_host_memory\x18\n \x01(\x04\x12\x15\n\rdevice_memory\x18\x0b \x01(\x04\x12\x1a\n\x12self_device_memory\x18\x0c \x01(\x04\x12\x11\n\tdata_size\x18\r \x01(\x04\x12\x14\n\x0c\x64\x61ta_per_sec\x18\x0e \x01(\x01\x12\r\n\x05\x66lops\x18\x0f \x01(\x04\"L\n\x06OpType\x12\x10\n\x0cUNDEFINED_OP\x10\x00\x12\r\n\tPYTHON_OP\x10\x01\x12\x0e\n\nPYTORCH_OP\x10\x02\x12\x11\n\rCOLLECTIVE_OP\x10\x03\"k\n\x0bKernelStats\x12\x12\n\ndevice_idx\x18\x01 \x01(\x04\x12\x0f\n\x07op_name\x18\x02 \x01(\t\x12\x13\n\x0bkernel_name\x18\x03 \x01(\t\x12\r\n\x05\x63ount\x18\x04 \x01(\x04\x12\x13\n\x0b\x64uration_ns\x18\x05 \x01(\x04\"\xa4\x01\n\tDataStats\x12\x11\n\tdata_name\x18\x01 \x01(\t\x12\x11\n\tdata_type\x18\x02 \x01(\t\x12\r\n\x05shape\x18\x04 \x03(\x04\x12\x38\n\x06\x63ounts\x18\x05 \x03(\x0b\x32(.graphsignal.signals.DataStats.DataCount\x1a(\n\tDataCount\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05\x63ount\x18\x02 \x01(\x04\"L\n\nDataSample\x12\x11\n\tdata_name\x18\x01 \x01(\t\x12\x14\n\x0c\x63ontent_type\x18\x02 \x01(\t\x12\x15\n\rcontent_bytes\x18\x03 \x01(\x0c\"O\n\tModelInfo\x12\x14\n\x0cmodel_format\x18\x04 \x01(\t\x12\x12\n\nmodel_name\x18\x03 \x01(\t\x12\x18\n\x10model_size_bytes\x18\x02 \x01(\x04\"\xbd\x01\n\rFrameworkInfo\x12\x0c\n\x04name\x18\x04 \x01(\t\x12,\n\x07version\x18\x02 \x01(\x0b\x32\x1b.graphsignal.signals.SemVer\x12\x41\n\x06params\x18\x03 \x03(\x0b\x32\x31.graphsignal.signals.FrameworkInfo.FrameworkParam\x1a-\n\x0e\x46rameworkParam\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"\xd3\x05\n\x0b\x44\x65viceUsage\x12@\n\x0b\x64\x65vice_type\x18\x0b \x01(\x0e\x32+.graphsignal.signals.DeviceUsage.DeviceType\x12\x12\n\ndevice_idx\x18\x15 \x01(\x04\x12\x11\n\tdevice_id\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65vice_name\x18\x02 \x01(\t\x12\x14\n\x0c\x61rchitecture\x18\x11 \x01(\t\x12\x37\n\x12\x63ompute_capability\x18\x0f \x01(\x0b\x32\x1b.graphsignal.signals.SemVer\x12\x11\n\tmem_total\x18\x03 \x01(\x04\x12\x10\n\x08mem_used\x18\x04 \x01(\x04\x12\x10\n\x08mem_free\x18\x05 \x01(\x04\x12\x14\n\x0cmem_reserved\x18\x13 \x01(\x04\x12\x1f\n\x17gpu_utilization_percent\x18\x06 \x01(\x01\x12\x1a\n\x12mem_access_percent\x18\x10 \x01(\x01\x12\x1a\n\x12pcie_throughput_tx\x18\r \x01(\x01\x12\x1a\n\x12pcie_throughput_rx\x18\x0e \x01(\x01\x12&\n\x1envlink_throughput_data_tx_kibs\x18\x17 \x01(\x01\x12&\n\x1envlink_throughput_data_rx_kibs\x18\x18 \x01(\x01\x12\x12\n\ngpu_temp_c\x18\x08 \x01(\x01\x12\x15\n\rpower_usage_w\x18\t \x01(\x01\x12\x19\n\x11\x66\x61n_speed_percent\x18\n \x01(\x01\x12\x1f\n\x17mxu_utilization_percent\x18\x0c \x01(\x01\x12:\n\tprocesses\x18\x14 \x03(\x0b\x32\'.graphsignal.signals.DeviceProcessUsage\"B\n\nDeviceType\x12\x14\n\x10UNDEFINED_DEVICE\x10\x00\x12\x0e\n\nGPU_DEVICE\x10\x02\x12\x0e\n\nTPU_DEVICE\x10\x03\"i\n\x12\x44\x65viceProcessUsage\x12\x0b\n\x03pid\x18\x01 \x01(\x04\x12\x17\n\x0fgpu_instance_id\x18\x02 \x01(\x04\x12\x1b\n\x13\x63ompute_instance_id\x18\x03 \x01(\x04\x12\x10\n\x08mem_used\x18\x04 \x01(\x04\"\x93\x03\n\x0cProcessUsage\x12\x0b\n\x03pid\x18\x13 \x01(\x04\x12\x0c\n\x04rank\x18\x14 \x01(\x04\x12\x10\n\x08has_rank\x18\x15 \x01(\x08\x12\x12\n\nlocal_rank\x18\x16 \x01(\x04\x12\x16\n\x0ehas_local_rank\x18\x17 \x01(\x08\x12\x10\n\x08start_ms\x18\x0e \x01(\x04\x12\x10\n\x08\x63pu_name\x18\x10 \x01(\t\x12\x19\n\x11\x63pu_usage_percent\x18\x01 \x01(\x01\x12\x0f\n\x07max_rss\x18\x02 \x01(\x04\x12\x13\n\x0b\x63urrent_rss\x18\x03 \x01(\x04\x12\x0f\n\x07vm_size\x18\x04 \x01(\x04\x12:\n\x07runtime\x18\x06 \x01(\x0e\x32).graphsignal.signals.ProcessUsage.Runtime\x12\x34\n\x0fruntime_version\x18\x07 \x01(\x0b\x32\x1b.graphsignal.signals.SemVer\x12\x14\n\x0cruntime_impl\x18\x08 \x01(\t\",\n\x07Runtime\x12\x15\n\x11RUNTIME_UNDEFINED\x10\x00\x12\n\n\x06PYTHON\x10\x01\"\xc6\x02\n\tNodeUsage\x12\x10\n\x08hostname\x18\x01 \x01(\t\x12\x12\n\nip_address\x18\x07 \x01(\t\x12\x11\n\tnode_rank\x18\x0c \x01(\x04\x12\x15\n\rhas_node_rank\x18\r \x01(\x08\x12\x10\n\x08mem_used\x18\t \x01(\x04\x12\x11\n\tmem_total\x18\n \x01(\x04\x12\x10\n\x08platform\x18\x02 \x01(\t\x12\x0f\n\x07machine\x18\x03 \x01(\t\x12\x0f\n\x07os_name\x18\x04 \x01(\t\x12\x12\n\nos_version\x18\x05 \x01(\t\x12\x13\n\x0bnum_devices\x18\x06 \x01(\x05\x12:\n\x07\x64rivers\x18\x0b \x03(\x0b\x32).graphsignal.signals.NodeUsage.DriverInfo\x1a+\n\nDriverInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\"\xb2\x01\n\nTracerInfo\x12?\n\x0btracer_type\x18\x01 \x01(\x0e\x32*.graphsignal.signals.TracerInfo.TracerType\x12,\n\x07version\x18\x03 \x01(\x0b\x32\x1b.graphsignal.signals.SemVer\"5\n\nTracerType\x12\x14\n\x10UNDEFINED_TRACER\x10\x00\x12\x11\n\rPYTHON_TRACER\x10\x01\"3\n\x0bTracerError\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x13\n\x0bstack_trace\x18\x02 \x01(\t\"\x91\x01\n\x0cMetricRecord\x12\x0f\n\x07\x64\x61ta_id\x18\x01 \x01(\t\x12\x11\n\tmetric_id\x18\x02 \x01(\t\x12+\n\x06metric\x18\x03 \x01(\x0b\x32\x1b.graphsignal.signals.Metric\x12\x1a\n\x12\x64\x61ta_retention_sec\x18\x04 \x01(\x04\x12\x14\n\x0ctime_skew_ms\x18\x05 \x01(\x12\"\xba\x03\n\x06Metric\x12\r\n\x05scope\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12&\n\x04tags\x18\x03 \x03(\x0b\x32\x18.graphsignal.signals.Tag\x12\x34\n\x04type\x18\x04 \x01(\x0e\x32&.graphsignal.signals.Metric.MetricType\x12\x0c\n\x04unit\x18\x05 \x01(\t\x12\x0f\n\x07is_time\x18\x06 \x01(\x08\x12\x0f\n\x07is_size\x18\x07 \x01(\x08\x12\x0f\n\x05gauge\x18\x08 \x01(\x01H\x00\x12\x11\n\x07\x63ounter\x18\t \x01(\x04H\x00\x12:\n\thistogram\x18\n \x01(\x0b\x32%.graphsignal.signals.Metric.HistogramH\x00\x12\x11\n\tupdate_ts\x18\x0b \x01(\x04\x1a)\n\tHistogram\x12\x0c\n\x04\x62ins\x18\x01 \x03(\x04\x12\x0e\n\x06\x63ounts\x18\x02 \x03(\x04\"^\n\nMetricType\x12\x14\n\x10UNDEFINED_METRIC\x10\x00\x12\x10\n\x0cGAUGE_METRIC\x10\x01\x12\x12\n\x0e\x43OUNTER_METRIC\x10\x02\x12\x14\n\x10HISTOGRAM_METRIC\x10\x03\x42\x07\n\x05valueb\x06proto3')



_UPLOADREQUEST = DESCRIPTOR.message_types_by_name['UploadRequest']
_UPLOADRESPONSE = DESCRIPTOR.message_types_by_name['UploadResponse']
_TRACERECORD = DESCRIPTOR.message_types_by_name['TraceRecord']
_TRACE = DESCRIPTOR.message_types_by_name['Trace']
_TAG = DESCRIPTOR.message_types_by_name['Tag']
_PARAM = DESCRIPTOR.message_types_by_name['Param']
_SPAN = DESCRIPTOR.message_types_by_name['Span']
_EXCEPTION = DESCRIPTOR.message_types_by_name['Exception']
_SEMVER = DESCRIPTOR.message_types_by_name['SemVer']
_MEMORYALLOCATION = DESCRIPTOR.message_types_by_name['MemoryAllocation']
_OPSTATS = DESCRIPTOR.message_types_by_name['OpStats']
_KERNELSTATS = DESCRIPTOR.message_types_by_name['KernelStats']
_DATASTATS = DESCRIPTOR.message_types_by_name['DataStats']
_DATASTATS_DATACOUNT = _DATASTATS.nested_types_by_name['DataCount']
_DATASAMPLE = DESCRIPTOR.message_types_by_name['DataSample']
_MODELINFO = DESCRIPTOR.message_types_by_name['ModelInfo']
_FRAMEWORKINFO = DESCRIPTOR.message_types_by_name['FrameworkInfo']
_FRAMEWORKINFO_FRAMEWORKPARAM = _FRAMEWORKINFO.nested_types_by_name['FrameworkParam']
_DEVICEUSAGE = DESCRIPTOR.message_types_by_name['DeviceUsage']
_DEVICEPROCESSUSAGE = DESCRIPTOR.message_types_by_name['DeviceProcessUsage']
_PROCESSUSAGE = DESCRIPTOR.message_types_by_name['ProcessUsage']
_NODEUSAGE = DESCRIPTOR.message_types_by_name['NodeUsage']
_NODEUSAGE_DRIVERINFO = _NODEUSAGE.nested_types_by_name['DriverInfo']
_TRACERINFO = DESCRIPTOR.message_types_by_name['TracerInfo']
_TRACERERROR = DESCRIPTOR.message_types_by_name['TracerError']
_METRICRECORD = DESCRIPTOR.message_types_by_name['MetricRecord']
_METRIC = DESCRIPTOR.message_types_by_name['Metric']
_METRIC_HISTOGRAM = _METRIC.nested_types_by_name['Histogram']
_TRACE_SAMPLINGTYPE = _TRACE.enum_types_by_name['SamplingType']
_MEMORYALLOCATION_ALLOCATORTYPE = _MEMORYALLOCATION.enum_types_by_name['AllocatorType']
_OPSTATS_OPTYPE = _OPSTATS.enum_types_by_name['OpType']
_DEVICEUSAGE_DEVICETYPE = _DEVICEUSAGE.enum_types_by_name['DeviceType']
_PROCESSUSAGE_RUNTIME = _PROCESSUSAGE.enum_types_by_name['Runtime']
_TRACERINFO_TRACERTYPE = _TRACERINFO.enum_types_by_name['TracerType']
_METRIC_METRICTYPE = _METRIC.enum_types_by_name['MetricType']
UploadRequest = _reflection.GeneratedProtocolMessageType('UploadRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPLOADREQUEST,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.UploadRequest)
  })
_sym_db.RegisterMessage(UploadRequest)

UploadResponse = _reflection.GeneratedProtocolMessageType('UploadResponse', (_message.Message,), {
  'DESCRIPTOR' : _UPLOADRESPONSE,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.UploadResponse)
  })
_sym_db.RegisterMessage(UploadResponse)

TraceRecord = _reflection.GeneratedProtocolMessageType('TraceRecord', (_message.Message,), {
  'DESCRIPTOR' : _TRACERECORD,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.TraceRecord)
  })
_sym_db.RegisterMessage(TraceRecord)

Trace = _reflection.GeneratedProtocolMessageType('Trace', (_message.Message,), {
  'DESCRIPTOR' : _TRACE,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.Trace)
  })
_sym_db.RegisterMessage(Trace)

Tag = _reflection.GeneratedProtocolMessageType('Tag', (_message.Message,), {
  'DESCRIPTOR' : _TAG,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.Tag)
  })
_sym_db.RegisterMessage(Tag)

Param = _reflection.GeneratedProtocolMessageType('Param', (_message.Message,), {
  'DESCRIPTOR' : _PARAM,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.Param)
  })
_sym_db.RegisterMessage(Param)

Span = _reflection.GeneratedProtocolMessageType('Span', (_message.Message,), {
  'DESCRIPTOR' : _SPAN,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.Span)
  })
_sym_db.RegisterMessage(Span)

Exception = _reflection.GeneratedProtocolMessageType('Exception', (_message.Message,), {
  'DESCRIPTOR' : _EXCEPTION,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.Exception)
  })
_sym_db.RegisterMessage(Exception)

SemVer = _reflection.GeneratedProtocolMessageType('SemVer', (_message.Message,), {
  'DESCRIPTOR' : _SEMVER,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.SemVer)
  })
_sym_db.RegisterMessage(SemVer)

MemoryAllocation = _reflection.GeneratedProtocolMessageType('MemoryAllocation', (_message.Message,), {
  'DESCRIPTOR' : _MEMORYALLOCATION,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.MemoryAllocation)
  })
_sym_db.RegisterMessage(MemoryAllocation)

OpStats = _reflection.GeneratedProtocolMessageType('OpStats', (_message.Message,), {
  'DESCRIPTOR' : _OPSTATS,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.OpStats)
  })
_sym_db.RegisterMessage(OpStats)

KernelStats = _reflection.GeneratedProtocolMessageType('KernelStats', (_message.Message,), {
  'DESCRIPTOR' : _KERNELSTATS,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.KernelStats)
  })
_sym_db.RegisterMessage(KernelStats)

DataStats = _reflection.GeneratedProtocolMessageType('DataStats', (_message.Message,), {

  'DataCount' : _reflection.GeneratedProtocolMessageType('DataCount', (_message.Message,), {
    'DESCRIPTOR' : _DATASTATS_DATACOUNT,
    '__module__' : 'signals_pb2'
    # @@protoc_insertion_point(class_scope:graphsignal.signals.DataStats.DataCount)
    })
  ,
  'DESCRIPTOR' : _DATASTATS,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.DataStats)
  })
_sym_db.RegisterMessage(DataStats)
_sym_db.RegisterMessage(DataStats.DataCount)

DataSample = _reflection.GeneratedProtocolMessageType('DataSample', (_message.Message,), {
  'DESCRIPTOR' : _DATASAMPLE,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.DataSample)
  })
_sym_db.RegisterMessage(DataSample)

ModelInfo = _reflection.GeneratedProtocolMessageType('ModelInfo', (_message.Message,), {
  'DESCRIPTOR' : _MODELINFO,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.ModelInfo)
  })
_sym_db.RegisterMessage(ModelInfo)

FrameworkInfo = _reflection.GeneratedProtocolMessageType('FrameworkInfo', (_message.Message,), {

  'FrameworkParam' : _reflection.GeneratedProtocolMessageType('FrameworkParam', (_message.Message,), {
    'DESCRIPTOR' : _FRAMEWORKINFO_FRAMEWORKPARAM,
    '__module__' : 'signals_pb2'
    # @@protoc_insertion_point(class_scope:graphsignal.signals.FrameworkInfo.FrameworkParam)
    })
  ,
  'DESCRIPTOR' : _FRAMEWORKINFO,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.FrameworkInfo)
  })
_sym_db.RegisterMessage(FrameworkInfo)
_sym_db.RegisterMessage(FrameworkInfo.FrameworkParam)

DeviceUsage = _reflection.GeneratedProtocolMessageType('DeviceUsage', (_message.Message,), {
  'DESCRIPTOR' : _DEVICEUSAGE,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.DeviceUsage)
  })
_sym_db.RegisterMessage(DeviceUsage)

DeviceProcessUsage = _reflection.GeneratedProtocolMessageType('DeviceProcessUsage', (_message.Message,), {
  'DESCRIPTOR' : _DEVICEPROCESSUSAGE,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.DeviceProcessUsage)
  })
_sym_db.RegisterMessage(DeviceProcessUsage)

ProcessUsage = _reflection.GeneratedProtocolMessageType('ProcessUsage', (_message.Message,), {
  'DESCRIPTOR' : _PROCESSUSAGE,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.ProcessUsage)
  })
_sym_db.RegisterMessage(ProcessUsage)

NodeUsage = _reflection.GeneratedProtocolMessageType('NodeUsage', (_message.Message,), {

  'DriverInfo' : _reflection.GeneratedProtocolMessageType('DriverInfo', (_message.Message,), {
    'DESCRIPTOR' : _NODEUSAGE_DRIVERINFO,
    '__module__' : 'signals_pb2'
    # @@protoc_insertion_point(class_scope:graphsignal.signals.NodeUsage.DriverInfo)
    })
  ,
  'DESCRIPTOR' : _NODEUSAGE,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.NodeUsage)
  })
_sym_db.RegisterMessage(NodeUsage)
_sym_db.RegisterMessage(NodeUsage.DriverInfo)

TracerInfo = _reflection.GeneratedProtocolMessageType('TracerInfo', (_message.Message,), {
  'DESCRIPTOR' : _TRACERINFO,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.TracerInfo)
  })
_sym_db.RegisterMessage(TracerInfo)

TracerError = _reflection.GeneratedProtocolMessageType('TracerError', (_message.Message,), {
  'DESCRIPTOR' : _TRACERERROR,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.TracerError)
  })
_sym_db.RegisterMessage(TracerError)

MetricRecord = _reflection.GeneratedProtocolMessageType('MetricRecord', (_message.Message,), {
  'DESCRIPTOR' : _METRICRECORD,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.MetricRecord)
  })
_sym_db.RegisterMessage(MetricRecord)

Metric = _reflection.GeneratedProtocolMessageType('Metric', (_message.Message,), {

  'Histogram' : _reflection.GeneratedProtocolMessageType('Histogram', (_message.Message,), {
    'DESCRIPTOR' : _METRIC_HISTOGRAM,
    '__module__' : 'signals_pb2'
    # @@protoc_insertion_point(class_scope:graphsignal.signals.Metric.Histogram)
    })
  ,
  'DESCRIPTOR' : _METRIC,
  '__module__' : 'signals_pb2'
  # @@protoc_insertion_point(class_scope:graphsignal.signals.Metric)
  })
_sym_db.RegisterMessage(Metric)
_sym_db.RegisterMessage(Metric.Histogram)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _UPLOADREQUEST._serialized_start=38
  _UPLOADREQUEST._serialized_end=162
  _UPLOADRESPONSE._serialized_start=164
  _UPLOADRESPONSE._serialized_end=180
  _TRACERECORD._serialized_start=183
  _TRACERECORD._serialized_end=327
  _TRACE._serialized_start=330
  _TRACE._serialized_end=1412
  _TRACE_SAMPLINGTYPE._serialized_start=1312
  _TRACE_SAMPLINGTYPE._serialized_end=1412
  _TAG._serialized_start=1414
  _TAG._serialized_end=1447
  _PARAM._serialized_start=1449
  _PARAM._serialized_end=1485
  _SPAN._serialized_start=1488
  _SPAN._serialized_end=1650
  _EXCEPTION._serialized_start=1652
  _EXCEPTION._serialized_end=1719
  _SEMVER._serialized_start=1721
  _SEMVER._serialized_end=1774
  _MEMORYALLOCATION._serialized_start=1777
  _MEMORYALLOCATION._serialized_end=2099
  _MEMORYALLOCATION_ALLOCATORTYPE._serialized_start=2031
  _MEMORYALLOCATION_ALLOCATORTYPE._serialized_end=2099
  _OPSTATS._serialized_start=2102
  _OPSTATS._serialized_end=2563
  _OPSTATS_OPTYPE._serialized_start=2487
  _OPSTATS_OPTYPE._serialized_end=2563
  _KERNELSTATS._serialized_start=2565
  _KERNELSTATS._serialized_end=2672
  _DATASTATS._serialized_start=2675
  _DATASTATS._serialized_end=2839
  _DATASTATS_DATACOUNT._serialized_start=2799
  _DATASTATS_DATACOUNT._serialized_end=2839
  _DATASAMPLE._serialized_start=2841
  _DATASAMPLE._serialized_end=2917
  _MODELINFO._serialized_start=2919
  _MODELINFO._serialized_end=2998
  _FRAMEWORKINFO._serialized_start=3001
  _FRAMEWORKINFO._serialized_end=3190
  _FRAMEWORKINFO_FRAMEWORKPARAM._serialized_start=3145
  _FRAMEWORKINFO_FRAMEWORKPARAM._serialized_end=3190
  _DEVICEUSAGE._serialized_start=3193
  _DEVICEUSAGE._serialized_end=3916
  _DEVICEUSAGE_DEVICETYPE._serialized_start=3850
  _DEVICEUSAGE_DEVICETYPE._serialized_end=3916
  _DEVICEPROCESSUSAGE._serialized_start=3918
  _DEVICEPROCESSUSAGE._serialized_end=4023
  _PROCESSUSAGE._serialized_start=4026
  _PROCESSUSAGE._serialized_end=4429
  _PROCESSUSAGE_RUNTIME._serialized_start=4385
  _PROCESSUSAGE_RUNTIME._serialized_end=4429
  _NODEUSAGE._serialized_start=4432
  _NODEUSAGE._serialized_end=4758
  _NODEUSAGE_DRIVERINFO._serialized_start=4715
  _NODEUSAGE_DRIVERINFO._serialized_end=4758
  _TRACERINFO._serialized_start=4761
  _TRACERINFO._serialized_end=4939
  _TRACERINFO_TRACERTYPE._serialized_start=4886
  _TRACERINFO_TRACERTYPE._serialized_end=4939
  _TRACERERROR._serialized_start=4941
  _TRACERERROR._serialized_end=4992
  _METRICRECORD._serialized_start=4995
  _METRICRECORD._serialized_end=5140
  _METRIC._serialized_start=5143
  _METRIC._serialized_end=5585
  _METRIC_HISTOGRAM._serialized_start=5439
  _METRIC_HISTOGRAM._serialized_end=5480
  _METRIC_METRICTYPE._serialized_start=5482
  _METRIC_METRICTYPE._serialized_end=5576
# @@protoc_insertion_point(module_scope)



import unittest
import logging
import sys
from unittest.mock import patch, Mock

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.proto_utils import parse_semver, compare_semver

logger = logging.getLogger('graphsignal')


class SystemInfoTest(unittest.TestCase):
    def setUp(self):
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            upload_on_shutdown=False,
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_parse_semver(self):
        s1 = signals_pb2.SemVer()
        parse_semver(s1, '1.2.3')
        self.assertEqual(s1.major, 1)
        self.assertEqual(s1.minor, 2)
        self.assertEqual(s1.patch, 3)

        s2 = signals_pb2.SemVer()
        parse_semver(s2, '1.2')
        self.assertEqual(s2.major, 1)
        self.assertEqual(s2.minor, 2)
        self.assertEqual(s2.patch, 0)

    def test_compare_semver(self):
        s1 = signals_pb2.SemVer()

        s1.major = 1
        s1.minor = 2
        self.assertEqual(compare_semver(s1, (1, 3, 0)), -1)

        s1.major = 1
        s1.minor = 2
        s1.patch = 3
        self.assertEqual(compare_semver(s1, (1, 2, 3)), 0)

        s1.major = 1
        s1.minor = 2
        s1.patch = 3
        self.assertEqual(compare_semver(s1, (1, 2, 2)), 1)



import unittest
import logging
import sys
import os
import time
from unittest.mock import patch, Mock
from google.protobuf.json_format import MessageToJson
import pprint
import random

import graphsignal
from graphsignal.proto import signals_pb2

logger = logging.getLogger('graphsignal')

class MetricStoreTest(unittest.TestCase):
    def setUp(self):
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            upload_on_shutdown=False,
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    @patch('time.time', return_value=1)
    def test_update_and_export(self, mocked_time):
        store = graphsignal._agent.metric_store()
        store.set_gauge(scope='s1', name='m1', tags={'t1': '1'}, value=1, update_ts=10, unit='u1', is_time=True, is_size=True)
        store.set_gauge(scope='s1', name='m1', tags={'t1': '1'}, value=2, update_ts=10, unit='u1', is_time=True, is_size=True)
        protos = store.export()
        self.assertEqual(len(protos), 1)
        self.assertEqual(protos[0].scope, 's1')
        self.assertEqual(protos[0].name, 'm1')
        self.assertEqual(protos[0].tags[0].key, 't1')
        self.assertEqual(protos[0].tags[0].value, '1')
        self.assertEqual(protos[0].unit, 'u1')
        self.assertEqual(protos[0].is_time, True)
        self.assertEqual(protos[0].is_size, True)
        self.assertEqual(protos[0].gauge, 2)
        self.assertEqual(protos[0].update_ts, 10)

        store.clear()

        store.inc_counter(scope='s1', name='m1', tags={'t1': '1'}, value=1, update_ts=10, unit='u1')
        store.inc_counter(scope='s1', name='m1', tags={'t1': '1'}, value=2, update_ts=10, unit='u1')
        protos = store.export()
        self.assertEqual(len(protos), 1)
        self.assertEqual(protos[0].scope, 's1')
        self.assertEqual(protos[0].name, 'm1')
        self.assertEqual(protos[0].tags[0].key, 't1')
        self.assertEqual(protos[0].tags[0].value, '1')
        self.assertEqual(protos[0].unit, 'u1')
        self.assertEqual(protos[0].is_time, False)
        self.assertEqual(protos[0].is_size, False)
        self.assertEqual(protos[0].counter, 3)
        self.assertEqual(protos[0].update_ts, 10)

        store.clear()

        store.update_histogram(scope='s1', name='m1', tags={'t1': '1'}, value=1, update_ts=10, unit='u1', is_time=True, is_size=True)
        store.update_histogram(scope='s1', name='m1', tags={'t1': '1'}, value=2, update_ts=10, unit='u1', is_time=True, is_size=True)
        protos = store.export()
        self.assertEqual(len(protos), 1)
        self.assertEqual(protos[0].scope, 's1')
        self.assertEqual(protos[0].name, 'm1')
        self.assertEqual(protos[0].tags[0].key, 't1')
        self.assertEqual(protos[0].tags[0].value, '1')
        self.assertEqual(protos[0].unit, 'u1')
        self.assertEqual(protos[0].is_time, True)
        self.assertEqual(protos[0].is_size, True)
        self.assertEqual(protos[0].histogram.bins, [1, 2])
        self.assertEqual(protos[0].histogram.counts, [1, 1])
        self.assertEqual(protos[0].update_ts, 10)

    def test_has_unexported(self):
        store = graphsignal._agent.metric_store()
        self.assertFalse(store.has_unexported())
        store.set_gauge(scope='s1', name='m1', tags={'t1': '1'}, value=1, update_ts=10, unit='u1', is_time=True, is_size=True)
        self.assertTrue(store.has_unexported())



import logging
import sys
import torch
import torch.distributed as dist

import graphsignal
from graphsignal.recorders.base_recorder import BaseRecorder
from graphsignal.proto_utils import parse_semver
from graphsignal.proto import signals_pb2
from graphsignal.proto_utils import add_framework_param, add_driver

logger = logging.getLogger('graphsignal')

class PyTorchRecorder(BaseRecorder):
    def __init__(self):
        self._framework = None
        self._comm_info = None
        self._rank = None
        self._is_cuda_available = False

    def setup(self):
        self._framework = signals_pb2.FrameworkInfo()
        self._framework.name = 'PyTorch'
        parse_semver(self._framework.version, torch.__version__)

        add_framework_param(self._framework, 'torch.cuda.is_available', torch.cuda.is_available())
        add_framework_param(self._framework, 'torch.backends.cuda.is_build', torch.backends.cuda.is_built())
        add_framework_param(self._framework, 'torch.backends.cudnn.is_available', torch.backends.cudnn.is_available())
        if torch.backends.cudnn.is_available(): 
            try:
                add_framework_param(self._framework, 'torch.backends.cudnn.is_available', _format_version(torch.backends.cudnn.version()))
            except RuntimeError:
                pass
        if hasattr(torch.backends, 'mps'):
            add_framework_param(self._framework, 'torch.backends.mps.is_available', torch.backends.mps.is_available())
            add_framework_param(self._framework, 'torch.backends.mps.is_built', torch.backends.mps.is_built())
        if hasattr(torch.backends, 'mkl'):
            add_framework_param(self._framework, 'torch.backends.mkl.is_available', torch.backends.mkl.is_available())
        if hasattr(torch.backends, 'mkldnn'):
            add_framework_param(self._framework, 'torch.backends.mkldnn.is_available', torch.backends.mkldnn.is_available())
        if hasattr(torch.backends, 'openmp'):
            add_framework_param(self._framework, 'torch.backends.openmp.is_available', torch.backends.openmp.is_available())
        add_framework_param(self._framework, 'torch.distributed.is_available', torch.distributed.is_available())
        if dist.is_available():
            add_framework_param(self._framework, 'torch.distributed.is_mpi_available', torch.distributed.is_mpi_available())
            add_framework_param(self._framework, 'torch.distributed.is_nccl_available', torch.distributed.is_nccl_available())
            add_framework_param(self._framework, 'torch.distributed.is_initialized', torch.distributed.is_initialized())
            if dist.is_initialized():
                add_framework_param(self._framework, 'torch.distributed.get_backend', torch.distributed.get_backend())
                add_framework_param(self._framework, 'torch.distributed.get_world_size', torch.distributed.get_world_size())
                add_framework_param(self._framework, 'torch.distributed.get_rank', torch.distributed.get_rank())
                self._rank = torch.distributed.get_rank()

        if torch.cuda.is_available():
            self._is_cuda_available = True

    def on_trace_start(self, proto, context, options):
        if self._is_cuda_available:
            context['pytorch_mem_stats'] = {}
            for device in range(torch.cuda.device_count()):
                context['pytorch_mem_stats'][device] = _read_mem_stats(device)

    def on_trace_stop(self, proto, context, options):
        if self._is_cuda_available:
            for device in range(torch.cuda.device_count()):
                if 'pytorch_mem_stats' in context and device in context['pytorch_mem_stats']: 
                    start_mem_stats = context['pytorch_mem_stats'][device]
                    stop_mem_stats = _read_mem_stats(device)

                    mem_diff = _compute_diff(start_mem_stats, stop_mem_stats)
                    mem_alloc = proto.alloc_summary.add()
                    mem_alloc.allocator_type = signals_pb2.MemoryAllocation.AllocatorType.PYTORCH_CUDA_ALLOCATOR
                    mem_alloc.device_idx = device
                    mem_alloc.allocated_size = mem_diff.get('allocated_size', 0)
                    mem_alloc.reserved_size = mem_diff.get('reserved_size', 0)
                    mem_alloc.freed_size = mem_diff.get('freed_size', 0)
                    mem_alloc.num_allocations = mem_diff.get('num_allocations', 0)
                    mem_alloc.num_alloc_retries = mem_diff.get('num_alloc_retries', 0)
                    mem_alloc.num_ooms = mem_diff.get('num_ooms', 0)

    def on_trace_read(self, proto, context, options):
        if self._framework:
            proto.frameworks.append(self._framework)
        if self._rank is not None:
            proto.process_usage.rank = self._rank
            proto.process_usage.has_rank = True

    def on_metric_update(self):
        pass

def _format_version(version):
    major = int(version / 1000)
    minor = int(version % 1000 / 100)
    patch = int(version % 10)
    return '{0}.{1}.{2}'.format(major, minor, patch)


def _read_mem_stats(device):
    mem_stats = torch.cuda.memory_stats(device)

    return dict(
        allocated_size=mem_stats.get("allocated_bytes.all.allocated", 0),
        reserved_size=mem_stats.get("allocated_bytes.all.reserved", 0),
        freed_size=mem_stats.get("allocated_bytes.all.freed", 0),
        num_allocations=mem_stats.get("allocation.all.allocated", 0),
        num_alloc_retries=mem_stats.get("num_alloc_retries", 0),
        num_ooms=mem_stats.get("num_ooms", 0)
    )

def _compute_diff(start_mem_stats, stop_mem_stats):
    diff = {}
    for key, stop_value in stop_mem_stats.items():
        start_value = start_mem_stats.get(key, 0)
        change = stop_value - start_value
        if change > 0:
            diff[key] = change
    return diff


import unittest
import logging
import sys
import os
import json
import time
from unittest.mock import patch, Mock
from google.protobuf.json_format import MessageToJson
import pprint

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.uploader import Uploader
from graphsignal.traces import DEFAULT_OPTIONS

logger = logging.getLogger('graphsignal')


class JAXRecorderTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            upload_on_shutdown=False,
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_record(self):
        try:
            import jax
            import jax.numpy as jnp
            from jax import grad, jit, vmap
            from jax import random
        except ImportError:
            logger.info('Not testing JAX, package not found.')
            return

        from graphsignal.recorders.jax_recorder import JAXRecorder

        recorder = JAXRecorder()
        recorder.setup()
        proto = signals_pb2.Trace()
        context = {}
        recorder.on_trace_start(proto, context, DEFAULT_OPTIONS)
        recorder.on_trace_stop(proto, context, DEFAULT_OPTIONS)
        recorder.on_trace_read(proto, context, DEFAULT_OPTIONS)

        self.assertEqual(proto.frameworks[0].name, 'JAX')



import logging
import os
import json
import tensorflow as tf

import graphsignal
from graphsignal.recorders.base_recorder import BaseRecorder
from graphsignal.proto_utils import parse_semver
from graphsignal.proto import signals_pb2
from graphsignal.proto_utils import add_framework_param

logger = logging.getLogger('graphsignal')

class TensorFlowRecorder(BaseRecorder):
    def __init__(self):
        self._framework = None
        self._comm_info = None

    def setup(self):
        self._framework = signals_pb2.FrameworkInfo()
        self._framework.name = 'TensorFlow'
        parse_semver(self._framework.version, tf.__version__)

        if 'TF_CONFIG' in os.environ:
            try:
                tf_config = json.loads(os.environ['TF_CONFIG'])

                cluster_size = 0
                if 'chief' in tf_config['cluster']:
                    cluster_size += len(tf_config['cluster']['chief'])
                if 'worker' in tf_config['cluster']:
                    cluster_size += len(tf_config['cluster']['worker'])
                if cluster_size > 0:
                    add_framework_param(self._framework, 'cluster_size', cluster_size)

                add_framework_param(self._framework, 'task_index', tf_config['task']['index'])
            except:
                logger.warning('Error parsing TF_CONFIG', exc_info=True)

        add_framework_param(self._framework, 'tf.test.is_built_with_gpu_support', tf.test.is_built_with_gpu_support())
        add_framework_param(self._framework, 'tf.test.is_built_with_cuda', tf.test.is_built_with_cuda())

    def on_trace_start(self, proto, context, options):
        pass

    def on_trace_stop(self, proto, context, options):
        pass

    def on_trace_read(self, proto, context, options):
        if self._framework:
            proto.frameworks.append(self._framework)

    def on_metric_update(self):
        pass



import unittest
import logging
import sys
import time
from unittest.mock import patch, Mock
import numpy as np
import tensorflow as tf

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.data.tf_tensor import TFTensorProfiler

logger = logging.getLogger('graphsignal')


class TFTensorProfilerTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            upload_on_shutdown=False,
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_is_instance(self):
        profiler = TFTensorProfiler()
        self.assertTrue(profiler.is_instance(tf.constant([])))

    def test_compute_stats(self):
        profiler = TFTensorProfiler()
        stats = profiler.compute_stats(tf.constant(np.asarray([[-1, 2.0, 0, np.inf], [1, 0, 0, np.nan]])))
        self.assertEqual(
            stats.counts, 
            {'element_count': 8,
             'byte_count': 64,
             'nan_count': 1,
             'inf_count': 1,
             'zero_count': 3,
             'negative_count': 1,
             'positive_count': 3})
        self.assertEqual(stats.type_name, 'tf.Tensor')
        self.assertEqual(stats.shape, [2, 4])



import unittest
import logging
import sys
import os
import json
import time
from unittest.mock import patch, Mock
from google.protobuf.json_format import MessageToJson
import pprint
import re

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.uploader import Uploader
from graphsignal.traces import DEFAULT_OPTIONS
from graphsignal.recorders.cprofile_recorder import CProfileRecorder, _format_frame

logger = logging.getLogger('graphsignal')


class CProfileRecorderTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            upload_on_shutdown=False,
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_record(self):
        recorder = CProfileRecorder()
        recorder._exclude_path = 'donotmatchpath'
        recorder.setup()
        proto = signals_pb2.Trace()
        context = {}

        def slow_method():
            time.sleep(0.1)

        recorder.on_trace_start(proto, context, graphsignal.TraceOptions(enable_profiling=True))
        slow_method()
        slow_method()
        recorder.on_trace_stop(proto, context, graphsignal.TraceOptions(enable_profiling=True))
        recorder.on_trace_read(proto, context, graphsignal.TraceOptions(enable_profiling=True))

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(proto))

        self.assertTrue('profiled' in proto.labels)

        slow_call = next(call for call in proto.op_profile if 'slow_method' in call.op_name)
        self.assertEqual(slow_call.op_type, signals_pb2.OpStats.OpType.PYTHON_OP)
        self.assertTrue(slow_call.self_host_time_ns > 0)
        self.assertTrue(slow_call.host_time_ns > 0)
        self.assertTrue(slow_call.self_host_time_percent > 0)

    def test_format_frame(self):
        self.assertEqual(_format_frame('p', 1, 'f'), 'f (p:1)')
        self.assertEqual(_format_frame('p', None, 'f'), 'f (p)')
        self.assertEqual(_format_frame(None, None, 'f'), 'f')
        self.assertEqual(_format_frame(None, None, None), 'unknown')


import unittest
import logging
import sys
import os
import json
import time
from unittest.mock import patch, Mock
from google.protobuf.json_format import MessageToJson
import pprint
import torch

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.uploader import Uploader
from graphsignal.traces import DEFAULT_OPTIONS
from graphsignal.recorders.kineto_recorder import KinetoRecorder

logger = logging.getLogger('graphsignal')


class KinetoRecorderTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            upload_on_shutdown=False,
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_record(self):
        recorder = KinetoRecorder()
        recorder.setup()

        x = torch.arange(-5, 5, 0.1).view(-1, 1)
        model = torch.nn.Linear(1, 1)
        if torch.cuda.is_available():
            x = x.to('cuda:0')
            model = model.to('cuda:0')

        proto = signals_pb2.Trace()
        context = {}
        recorder.on_trace_start(proto, context, graphsignal.TraceOptions(enable_profiling=True))

        pred = model(x)

        recorder.on_trace_stop(proto, context, graphsignal.TraceOptions(enable_profiling=True))
        recorder.on_trace_read(proto, context, graphsignal.TraceOptions(enable_profiling=True))

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(proto))

        self.assertTrue('profiled' in proto.labels)

        test_op_stats = None
        for op_stats in proto.op_profile:
            if op_stats.op_name == 'aten::addmm':
                test_op_stats = op_stats
                break
        self.assertIsNotNone(test_op_stats)
        self.assertTrue(test_op_stats.count >= 1)
        
        self.assertEqual(test_op_stats.op_type, signals_pb2.OpStats.OpType.PYTORCH_OP)
        if torch.cuda.is_available():
            self.assertTrue(test_op_stats.device_time_ns >= 1)
            self.assertTrue(test_op_stats.self_device_time_ns >= 1)
        else:
            self.assertTrue(test_op_stats.host_time_ns >= 1)
            self.assertTrue(test_op_stats.self_host_time_ns >= 1)
            self.assertTrue(test_op_stats.self_host_time_percent > 0)



import logging
import sys
import time
import banana_dev as banana

import graphsignal
from graphsignal.traces import TraceOptions
from graphsignal.recorders.base_recorder import BaseRecorder
from graphsignal.recorders.instrumentation import instrument_method, uninstrument_method, read_args
from graphsignal.proto_utils import parse_semver, compare_semver
from graphsignal.proto import signals_pb2
from graphsignal.proto_utils import add_framework_param, add_driver

logger = logging.getLogger('graphsignal')

class BananaRecorder(BaseRecorder):
    def __init__(self):
        self._framework = None

    def setup(self):
        if not graphsignal._agent.auto_instrument:
            return

        self._framework = signals_pb2.FrameworkInfo()
        self._framework.name = 'Banana Python SDK'

        instrument_method(banana, 'run', 'https://api.banana.dev', self.trace_run)

    def shutdown(self):
        uninstrument_method(banana, 'run', 'https://api.banana.dev')

    def trace_run(self, trace, args, kwargs, ret, exc):
        params = read_args(args, kwargs, ['api_key', 'model_key', 'model_inputs'])

        if 'model_key' in params:
            trace.set_tag('model_key', params['model_key'])
            trace.set_param('model_key', params['model_key'])

        if 'model_inputs' in params:
            trace.set_data('model_inputs', params['model_inputs'])

        if ret and 'modelOutputs' in ret:
            trace.set_data('model_outputs', ret['modelOutputs'])

    def on_trace_start(self, proto, context, options):
        pass

    def on_trace_stop(self, proto, context, options):
        pass

    def on_trace_read(self, proto, context, options):
        if self._framework:
            proto.frameworks.append(self._framework)

    def on_metric_update(self):
        pass


import logging
import sys
import threading
import time
import math

import graphsignal
from graphsignal import version
from graphsignal.proto import signals_pb2

logger = logging.getLogger('graphsignal')


class BaseMetric:
    def __init__(self, scope, name, tags, unit=None, is_time=False, is_size=False):
        self._update_lock = threading.Lock()
        self.scope = scope
        self.name = name
        self.tags = tags
        self.unit = unit
        self.is_time = is_time
        self.is_size = is_size
        self.is_updated = False

    def update(self):
        self.is_updated = True

    def export(self):
        self.is_updated = False

        proto = signals_pb2.Metric()
        proto.scope = self.scope
        proto.name = self.name
        for key, value in self.tags.items():
            tag = proto.tags.add()
            tag.key = str(key)[:50]
            tag.value = str(value)[:50]
        proto.type = self.type
        if self.unit is not None:
            proto.unit = self.unit
        proto.is_time = self.is_time
        proto.is_size = self.is_size
        proto.update_ts = self.update_ts
        return proto


class GaugeMetric(BaseMetric):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.type = signals_pb2.Metric.MetricType.GAUGE_METRIC
        self.gauge = None

    def update(self, value, update_ts):
        with self._update_lock:
            super().update()
            self.gauge = value
            self.update_ts = update_ts

    def export(self):
        with self._update_lock:
            proto = super().export()
            if self.gauge is not None:
                proto.gauge = self.gauge
            self.gauge = None
            return proto


class CounterMetric(BaseMetric):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.type = signals_pb2.Metric.MetricType.COUNTER_METRIC
        self.counter = 0

    def update(self, value, update_ts):
        with self._update_lock:
            super().update()
            self.counter += value
            self.update_ts = update_ts

    def export(self):
        with self._update_lock:
            proto = super().export()
            if self.counter > 0:
                proto.counter = self.counter
                self.counter = 0
            return proto


class HistogramMetric(BaseMetric):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.type = signals_pb2.Metric.MetricType.HISTOGRAM_METRIC
        self.histogram = {}

    def update(self, value, update_ts):
        if value < 1:
            return
        with self._update_lock:
            super().update()
            bin_size = max(10 ** (int(math.log(value, 10)) - 1), 1)
            bin = int(value / bin_size) * bin_size
            self.histogram[bin] = self.histogram.get(bin, 0) + 1
            self.update_ts = update_ts

    def export(self):
        with self._update_lock:
            proto = super().export()
            for bin, count in self.histogram.items():
                proto.histogram.bins.append(bin)
                proto.histogram.counts.append(count)
            self.histogram = {}
            return proto


class MetricStore:
    MIN_INTERVAL_SEC = 10

    def __init__(self):
        self._update_lock = threading.Lock()
        self._has_unexported = False
        self.metrics = {}

    def metric_key(self, scope, name, tags):
        return (scope, name, frozenset(tags.items()))

    def set_gauge(self, scope, name, tags, value, update_ts, unit=None, is_time=False, is_size=False):
        key = self.metric_key(scope, name, tags)
        with self._update_lock:
            if key not in self.metrics:
                metric = self.metrics[key] = GaugeMetric(scope, name, tags, unit=unit, is_time=is_time, is_size=is_size)
            else:
                metric = self.metrics[key]
        metric.update(value, update_ts)
        return self.metrics[key]

    def inc_counter(self, scope, name, tags, value, update_ts, unit=None):
        key = self.metric_key(scope, name, tags)
        with self._update_lock:
            if key not in self.metrics:
                metric = self.metrics[key] = CounterMetric(scope, name, tags, unit=unit)
            else:
                metric = self.metrics[key]
        metric.update(value, update_ts)
        return self.metrics[key]

    def update_histogram(self, scope, name, tags, value, update_ts, unit=None, is_time=False, is_size=False):
        key = self.metric_key(scope, name, tags)
        with self._update_lock:
            if key not in self.metrics:
                metric = self.metrics[key] = HistogramMetric(scope, name, tags, unit=unit, is_time=is_time, is_size=is_size)
            else:
                metric = self.metrics[key]
        metric.update(value, update_ts)
        return self.metrics[key]

    def has_unexported(self):
        with self._update_lock:
            for metric in self.metrics.values():
                if metric.is_updated:
                    return True
        return False

    def export(self):
        protos = []
        with self._update_lock:
            for metric in self.metrics.values():
                if metric.is_updated:
                    protos.append(metric.export())
        return protos

    def clear(self):
        with self._update_lock:
            self.metrics = {}


import unittest
import logging
import sys
import time
import random
import numpy as np
from unittest.mock import patch, Mock

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.traces import Trace, TraceOptions, DEFAULT_OPTIONS
from graphsignal.recorders.process_recorder import ProcessRecorder
from graphsignal.trace_sampler import TraceSampler
from graphsignal.uploader import Uploader


logger = logging.getLogger('graphsignal')


class TraceTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            tags={'k1': 'v1'},
            record_data_samples=True,
            upload_on_shutdown=False,
            debug_mode=True)
        graphsignal._agent.hostname = 'h1'

    def tearDown(self):
        graphsignal.shutdown()

    @patch.object(ProcessRecorder, 'on_trace_stop')
    @patch.object(Uploader, 'upload_trace')
    def test_start_stop(self, mocked_upload_trace, mocked_process_on_trace_stop):
        graphsignal.set_tag('k2', 'v2')

        graphsignal.set_context_tag('k3', 'v3')
        graphsignal.set_context_tag('k4', 'v4')

        for i in range(10):
            trace = Trace(
                endpoint='ep1',
                tags={'k4': 4.0})
            trace.set_tag('k5', 'v5')
            trace.set_param('p1', 'v1')
            trace.set_data('input', np.asarray([[1, 2],[3, 4]]), counts=dict(c1=1, c2=2))
            time.sleep(0.01)
            trace.stop()

        trace = mocked_upload_trace.call_args[0][0]

        self.assertTrue(trace.start_us > 0)
        self.assertTrue(trace.end_us > 0)
        self.assertEqual(trace.sampling_type, signals_pb2.Trace.SamplingType.RANDOM_SAMPLING)
        self.assertEqual(trace.labels, ['root'])
        self.assertTrue(trace.process_usage.start_ms > 0)
        self.assertEqual(trace.tags[0].key, 'deployment')
        self.assertEqual(trace.tags[0].value, 'd1')
        self.assertEqual(trace.tags[1].key, 'endpoint')
        self.assertEqual(trace.tags[1].value, 'ep1')
        self.assertEqual(trace.tags[2].key, 'hostname')
        self.assertIsNotNone(trace.tags[2].value)
        self.assertEqual(trace.tags[3].key, 'k1')
        self.assertEqual(trace.tags[3].value, 'v1')
        self.assertEqual(trace.tags[4].key, 'k2')
        self.assertEqual(trace.tags[4].value, 'v2')
        self.assertEqual(trace.tags[5].key, 'k3')
        self.assertEqual(trace.tags[5].value, 'v3')
        self.assertEqual(trace.tags[6].key, 'k4')
        self.assertEqual(trace.tags[6].value, '4.0')
        self.assertEqual(trace.tags[7].key, 'k5')
        self.assertEqual(trace.tags[7].value, 'v5')
        self.assertEqual(trace.params[0].name, 'p1')
        self.assertEqual(trace.params[0].value, 'v1')
        self.assertEqual(trace.data_profile[0].data_name, 'input')
        self.assertEqual(trace.data_profile[0].shape, [2, 2])
        self.assertEqual(trace.data_profile[0].counts[-2].name, 'c1')
        self.assertEqual(trace.data_profile[0].counts[-2].count, 1)
        self.assertEqual(trace.data_profile[0].counts[-1].name, 'c2')
        self.assertEqual(trace.data_profile[0].counts[-1].count, 2)
        self.assertEqual(trace.data_samples[0].data_name, 'input')
        self.assertEqual(trace.data_samples[0].content_type, 'application/json')
        self.assertEqual(trace.data_samples[0].content_bytes, b'[[1, 2], [3, 4]]')

        store = graphsignal._agent.metric_store()
        metric_tags =  {'deployment': 'd1', 'endpoint': 'ep1', 'hostname': 'h1', 'k1': 'v1', 'k2': 'v2', 'k3': 'v3', 'k4': 4.0, 'k5': 'v5'}
        key = store.metric_key('performance', 'latency', metric_tags)
        self.assertTrue(len(store.metrics[key].histogram) > 0)
        key = store.metric_key('performance', 'call_count', metric_tags)
        self.assertEqual(store.metrics[key].counter, 10)

        data_tags = metric_tags.copy()
        data_tags['data'] = 'input'
        key = store.metric_key('data', 'element_count', data_tags)
        self.assertEqual(store.metrics[key].counter, 40)
        key = store.metric_key('data', 'c1', data_tags)
        self.assertEqual(store.metrics[key].counter, 10)
        key = store.metric_key('data', 'c2', data_tags)
        self.assertEqual(store.metrics[key].counter, 20)

    @patch.object(ProcessRecorder, 'on_trace_start')
    @patch.object(Uploader, 'upload_trace')
    def test_start_stop_nested(self, mocked_upload_trace, mocked_process_on_trace_start):
        with Trace(endpoint='ep1') as trace1:
            with Trace(endpoint='ep2') as trace2:
                pass

        t1 = mocked_upload_trace.call_args_list[1][0][0]
        t2 = mocked_upload_trace.call_args_list[0][0][0]

        self.assertEqual(t1.sampling_type, signals_pb2.Trace.SamplingType.RANDOM_SAMPLING)
        self.assertTrue(t1.start_us > 0)
        self.assertTrue(t1.end_us > 0)
        self.assertEqual(t1.labels, ['root'])
        self.assertEqual(t2.span.parent_trace_id, t1.trace_id)
        self.assertEqual(t2.span.root_trace_id, t1.trace_id)

        self.assertEqual(t2.sampling_type, signals_pb2.Trace.SamplingType.PARENT_SAMPLING)
        self.assertTrue(t2.start_us > 0)
        self.assertTrue(t2.end_us > 0)
        self.assertEqual(t2.labels, [])


    @patch.object(ProcessRecorder, 'on_trace_start')
    @patch.object(Uploader, 'upload_trace')
    def test_start_exception(self, mocked_upload_trace, mocked_process_on_trace_start):
        mocked_process_on_trace_start.side_effect = Exception('ex1')
        trace = Trace(endpoint='ep1')
        trace.stop()

        trace = mocked_upload_trace.call_args[0][0]

        self.assertEqual(trace.sampling_type, signals_pb2.Trace.SamplingType.RANDOM_SAMPLING)
        self.assertTrue(trace.start_us > 0)
        self.assertTrue(trace.end_us > 0)
        self.assertEqual(trace.labels, ['root'])
        self.assertEqual(trace.tracer_errors[0].message, 'ex1')
        self.assertNotEqual(trace.tracer_errors[0].stack_trace, '')

    @patch.object(ProcessRecorder, 'on_trace_stop')
    @patch.object(Uploader, 'upload_trace')
    def test_agent_exception(self, mocked_upload_trace, mocked_process_on_trace_stop):
        mocked_process_on_trace_stop.side_effect = Exception('ex1')
        trace = Trace(
            endpoint='ep1')
        trace.stop()

        trace = mocked_upload_trace.call_args[0][0]

        self.assertTrue(trace.start_us > 0)
        self.assertTrue(trace.end_us > 0)
        self.assertEqual(trace.tracer_errors[0].message, 'ex1')
        self.assertNotEqual(trace.tracer_errors[0].stack_trace, '')

    @patch.object(Uploader, 'upload_trace')
    def test_endpoint_exception(self, mocked_upload_trace):
        for _ in range(2):
            try:
                with Trace(endpoint='ep1'):
                    raise Exception('ex1')
            except Exception as ex:
                if str(ex) != 'ex1':
                    raise ex

        self.assertEqual(mocked_upload_trace.call_count, 2)
        trace = mocked_upload_trace.call_args[0][0]

        self.assertEqual(trace.sampling_type, signals_pb2.Trace.SamplingType.RANDOM_SAMPLING)
        self.assertTrue(trace.trace_id != '')
        self.assertTrue(trace.start_us > 0)
        self.assertTrue(trace.end_us > 0)
        self.assertEqual(trace.labels, ['root', 'exception'])
        self.assertEqual(trace.exceptions[0].exc_type, 'Exception')
        self.assertEqual(trace.exceptions[0].message, 'ex1')
        self.assertNotEqual(trace.exceptions[0].stack_trace, '')

        store = graphsignal._agent.metric_store()
        metric_tags =  {'deployment': 'd1', 'endpoint': 'ep1', 'hostname': 'h1', 'k1': 'v1'}
        key = store.metric_key('performance', 'exception_count', metric_tags)
        self.assertEqual(store.metrics[key].counter, 2)

    @patch.object(Uploader, 'upload_trace')
    def test_set_exception(self, mocked_upload_trace):
        trace = Trace(endpoint='ep1')
        try:
            raise Exception('ex2')
        except Exception as ex:
            trace.set_exception(ex)
        trace.stop()

        trace = mocked_upload_trace.call_args[0][0]

        self.assertEqual(trace.sampling_type, signals_pb2.Trace.SamplingType.RANDOM_SAMPLING)
        self.assertTrue(trace.start_us > 0)
        self.assertTrue(trace.end_us > 0)
        self.assertEqual(trace.labels, ['root', 'exception'])
        self.assertEqual(trace.exceptions[0].exc_type, 'Exception')
        self.assertEqual(trace.exceptions[0].message, 'ex2')
        self.assertNotEqual(trace.exceptions[0].stack_trace, '')

        store = graphsignal._agent.metric_store()
        metric_tags =  {'deployment': 'd1', 'endpoint': 'ep1', 'hostname': 'h1', 'k1': 'v1'}
        key = store.metric_key('performance', 'exception_count', metric_tags)
        self.assertEqual(store.metrics[key].counter, 1)


    @patch.object(Uploader, 'upload_trace')
    def test_set_exception_true(self, mocked_upload_trace):
        trace = Trace(endpoint='ep1')
        try:
            raise Exception('ex2')
        except Exception as ex:
            trace.set_exception(exc_info=True)
        trace.stop()

        trace = mocked_upload_trace.call_args[0][0]

        self.assertEqual(trace.sampling_type, signals_pb2.Trace.SamplingType.RANDOM_SAMPLING)
        self.assertTrue(trace.start_us > 0)
        self.assertTrue(trace.end_us > 0)
        self.assertEqual(trace.labels, ['root', 'exception'])
        self.assertEqual(trace.exceptions[0].exc_type, 'Exception')
        self.assertEqual(trace.exceptions[0].message, 'ex2')
        self.assertNotEqual(trace.exceptions[0].stack_trace, '')

        store = graphsignal._agent.metric_store()
        metric_tags =  {'deployment': 'd1', 'endpoint': 'ep1', 'hostname': 'h1', 'k1': 'v1'}
        key = store.metric_key('performance', 'exception_count', metric_tags)
        self.assertEqual(store.metrics[key].counter, 1)

    @patch.object(Uploader, 'upload_trace')
    def test_outlier(self, mocked_upload_trace):
        for _ in range(500):
            with Trace(endpoint='ep1'):
                time.sleep(0.00001)
        with Trace(endpoint='ep1'):
            time.sleep(0.01)

        has_outliers = False
        for call_args in mocked_upload_trace.call_args_list:
            trace = call_args[0][0]
            if 'latency-outlier' in trace.labels:
                has_outliers = True
                self.assertEqual(trace.sampling_type, signals_pb2.Trace.SamplingType.ERROR_SAMPLING)
                break
        self.assertTrue(has_outliers)

    @patch.object(Uploader, 'upload_trace')
    def test_set_data(self, mocked_upload_trace):
        with Trace(endpoint='ep1') as trace:
            trace.set_data('d1', {'c1': 100, 'c2': None}, check_missing_values=True)

        trace = mocked_upload_trace.call_args[0][0]

        self.assertTrue(trace.start_us > 0)
        self.assertTrue(trace.end_us > 0)
        self.assertEqual(trace.labels, ['root', 'missing-values'])

    @patch.object(Uploader, 'upload_trace')
    def test_spans(self, mocked_upload_trace):
        with Trace(endpoint='ep1'):
            with Trace(endpoint='ep2'):
                with Trace(endpoint='ep3'):
                    pass
            with Trace(endpoint='ep4'):
                pass

        t3 = mocked_upload_trace.call_args_list[0][0][0]
        t2 = mocked_upload_trace.call_args_list[1][0][0]
        t4 = mocked_upload_trace.call_args_list[2][0][0]
        t1 = mocked_upload_trace.call_args_list[3][0][0]

        self.assertEqual(t1.span.parent_trace_id, '')
        self.assertEqual(t1.span.root_trace_id, t1.trace_id)

        self.assertEqual(t2.span.parent_trace_id, t1.trace_id)
        self.assertEqual(t2.span.root_trace_id, t1.trace_id)

        self.assertEqual(t3.span.parent_trace_id, t2.trace_id)
        self.assertEqual(t3.span.root_trace_id, t1.trace_id)

        self.assertEqual(t4.span.parent_trace_id, t1.trace_id)
        self.assertEqual(t4.span.root_trace_id, t1.trace_id)

    @patch.object(Uploader, 'upload_trace')
    @patch.object(TraceSampler, 'lock', return_value=False)
    def test_overhead(self, mocked_lock, mocked_upload_trace):
        #import cProfile, pstats
        #profiler = cProfile.Profile()
        #profiler.enable()

        graphsignal._agent.debug_mode = False

        calls = 10000
        start_ns = time.perf_counter_ns()
        for _ in range(calls):
            with Trace(endpoint='ep1') as trace:
                trace.set_tag('k1', 'v1')
                trace.set_param('k2', 'v2')
                trace.set_data('test', 'test string data', check_missing_values=True)
        took_ns = time.perf_counter_ns() - start_ns

        #stats = pstats.Stats(profiler).sort_stats('time')
        #stats.print_stats()

        self.assertTrue(took_ns / calls < 200 * 1e3) # less than 200 microseconds per trace



import logging
import sys
import torch
import torch.distributed as dist

import graphsignal
from graphsignal.recorders.base_recorder import BaseRecorder
from graphsignal.proto_utils import parse_semver
from graphsignal.proto import signals_pb2
from graphsignal.proto_utils import add_framework_param, add_driver

logger = logging.getLogger('graphsignal')

class KinetoRecorder(BaseRecorder):
    def __init__(self):
        self._torch_prof = None

    def setup(self):
        pass

    def on_trace_start(self, proto, context, options):
        if not options.enable_profiling:
            return

        if not self._torch_prof:
            def _schedule_func(step):
                return torch.profiler.ProfilerAction.RECORD
            self._torch_prof = torch.profiler.profile(
                schedule=_schedule_func,
                record_shapes=False,
                profile_memory=True,
                with_stack=False,
                with_flops=True)

        self._torch_prof.start()
        context['is_kineto_profiling'] = True

    def on_trace_stop(self, proto, context, options):
        if not context.get('is_kineto_profiling', False):
            return

        self._torch_prof.stop()

    def on_trace_read(self, proto, context, options):
        if not context.get('is_kineto_profiling', False):
            return

        total_self_host_time_ns = 0
        for event_avg in self._torch_prof.key_averages():
            if event_avg.key and event_avg.key.startswith('ProfilerStep'):
                continue
            op_stats = proto.op_profile.add()
            op_stats.op_type = signals_pb2.OpStats.OpType.PYTORCH_OP
            op_stats.op_name = event_avg.key
            op_stats.count = _uint(event_avg.count)
            op_stats.host_time_ns = _ns(event_avg.cpu_time_total)
            op_stats.device_time_ns = _ns(event_avg.cuda_time_total)
            op_stats.self_host_time_ns = _ns(event_avg.self_cpu_time_total)
            op_stats.self_device_time_ns = _ns(event_avg.self_cuda_time_total)
            op_stats.host_memory = _uint(event_avg.cpu_memory_usage)
            op_stats.device_memory = _uint(event_avg.cuda_memory_usage)
            op_stats.self_host_memory = _uint(event_avg.self_cpu_memory_usage)
            op_stats.self_device_memory = _uint(event_avg.self_cuda_memory_usage)
            op_stats.flops = _uint(event_avg.flops)
            total_self_host_time_ns += op_stats.self_host_time_ns

        if total_self_host_time_ns > 0:
            for op_stats in proto.op_profile:
                op_stats.self_host_time_percent = op_stats.self_host_time_ns / total_self_host_time_ns * 100

        kernel_index = {}
        for event in self._torch_prof.events():
            for kernel in event.kernels:
                key = (event.key, kernel.name, kernel.device)
                if key in kernel_index:
                    kernel_stats = kernel_index[key]
                    kernel_stats.count += 1
                    kernel_stats.duration_ns += _ns(kernel.duration)
                else:
                    kernel_stats = kernel_index[key] = signals_pb2.KernelStats()
                    kernel_stats.device_idx = kernel.device
                    kernel_stats.op_name = event.name
                    kernel_stats.kernel_name = kernel.name
                    kernel_stats.count = 1
                    kernel_stats.duration_ns = _ns(kernel.duration)

        for kernel_stats in kernel_index.values():
            proto.kernel_profile.append(kernel_stats)

        proto.labels.append('profiled')

    def on_metric_update(self):
        pass

def _ns(val):
    return int(max(val, 0) * 1e3)


def _uint(val):
    return max(int(val), 0)



import unittest
import logging
import sys
import time
from unittest.mock import patch, Mock
import numpy as np

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.data.numpy_ndarray import NumpyNDArrayProfiler

logger = logging.getLogger('graphsignal')


class NumpyNDArrayProfilerTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            upload_on_shutdown=False,
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_is_instance(self):
        profiler = NumpyNDArrayProfiler()
        self.assertTrue(profiler.is_instance(np.asarray([])))

    def test_compute_stats(self):
        profiler = NumpyNDArrayProfiler()
        stats = profiler.compute_stats(np.asarray([[-1, 2.0, 0, np.inf], [1, 0, 0, np.nan]]))
        self.assertEqual(
            stats.counts, 
            {'element_count': 8, 
            'byte_count': 64,
            'nan_count': 1, 
            'inf_count': 1, 
            'zero_count': 3,
            'negative_count': 1,
            'positive_count': 3})
        self.assertEqual(stats.type_name, 'numpy.ndarray')
        self.assertEqual(stats.shape, [2, 4])

    def test_encode_sample(self):
        profiler = NumpyNDArrayProfiler()
        preview = profiler.encode_sample(np.asarray([[-1, 2.0, 0, np.inf]]))
        self.assertEqual(preview.content_type, 'application/json')
        self.assertEqual(preview.content_bytes, b'[[-1.0, 2.0, 0.0, Infinity]]')



import logging
from functools import wraps
import asyncio
import types

import graphsignal

logger = logging.getLogger('graphsignal')


def instrument_method(obj, func_name, endpoint, trace_func, data_func=None):
    def before_func(args, kwargs):
        return dict(trace=graphsignal.start_trace(endpoint=endpoint))

    def after_func(args, kwargs, ret, exc, context):
        trace = context['trace']

        if not is_generator(ret):
            trace.measure()

        try:
            if exc is not None:
                trace.set_exception(exc)

            trace_func(trace, args, kwargs, ret, exc)
        except Exception as e:
            logger.debug('Error tracing %s', func_name, exc_info=True)

        if not is_generator(ret):
            trace.stop()

    def yield_func(idx, item, context):
        trace = context['trace']

        if idx == -1:
            trace.stop()
        else:
            if data_func:
                data_func(trace, item)

    if not patch_method(obj, func_name, before_func=before_func, after_func=after_func, yield_func=yield_func):
        logger.debug('Cannot instrument %s.', endpoint)


def uninstrument_method(obj, func_name, endpoint):
    if not unpatch_method(obj, func_name):
        logger.debug('Cannot uninstrument %s.', endpoint)


def patch_method(obj, func_name, before_func=None, after_func=None, yield_func=None):
    if not hasattr(obj, func_name):
        return False

    func = getattr(obj, func_name)

    if hasattr(func, '__graphsignal_wrapped__'):
        return False

    if yield_func:
        def gen_wrapper(gen, yield_func, context):
            for idx, item in enumerate(gen):
                try:
                    yield_func(idx, item, context)
                except:
                    logger.debug('Exception in yield_func', exc_info=True)
                yield item
            yield_func(-1, None, context)

    if asyncio.iscoroutinefunction(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            context = None
            exc = None
            ret = None

            if before_func:
                try:
                    context = before_func(args, kwargs)
                except:
                    logger.debug('Exception in before_func', exc_info=True)

            try:
                ret = await func(*args, **kwargs)
            except BaseException as e:
                exc = e

            if after_func:
                try:
                    after_func(args, kwargs, ret, exc, context)
                except:
                    logger.debug('Exception in after_func', exc_info=True)

            if yield_func:
                try:
                    if is_generator(ret):
                        ret = gen_wrapper(ret, yield_func, context)
                except:
                    logger.debug('Exception in yield_func', exc_info=True)

            if exc:
                raise exc
            return ret
    else:
        @wraps(func)
        def wrapper(*args, **kwargs):
            context = None
            exc = None
            ret = None

            if before_func:
                try:
                    context = before_func(args, kwargs)
                except:
                    logger.debug('Exception in before_func', exc_info=True)

            try:
                ret = func(*args, **kwargs)
            except BaseException as e:
                exc = e

            if after_func:
                try:
                    after_func(args, kwargs, ret, exc, context)
                except:
                    logger.debug('Exception in after_func', exc_info=True)

            if yield_func:
                try:
                    if is_generator(ret):
                        ret = gen_wrapper(ret, yield_func, context)
                except:
                    logger.debug('Exception in yield_func', exc_info=True)

            if exc:
                raise exc
            return ret

    setattr(wrapper, '__graphsignal_wrapped__', True)
    setattr(obj, func_name, wrapper)
    return True


def unpatch_method(obj, func_name):
    if not hasattr(obj, func_name):
        return False

    func = getattr(obj, func_name)

    if not hasattr(func, '__graphsignal_wrapped__'):
        return False

    if not hasattr(func, '__wrapped__'):
        return False
    
    setattr(obj, func_name, getattr(func, '__wrapped__'))
    return True


def is_generator(obj):
    return obj and isinstance(obj, types.GeneratorType)


def read_args(args, kwargs, names):
    values = {}
    for name, arg in zip(names, args):
        values[name] = arg
    values.update(kwargs)
    return values


import unittest
import logging
import sys
import time
from unittest.mock import patch, Mock
import asyncio
import random

import graphsignal
from graphsignal.spans import get_current_span, Span

logger = logging.getLogger('graphsignal')

class SpanTest(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            upload_on_shutdown=False,
            debug_mode=True)

    async def asyncTearDown(self):
        graphsignal.shutdown()

    async def test_span_nested(self):
        s1 = Span('s1')
        s1.set_trace_id('t1')
        s1.set_sampling(True)
        self.assertEqual(get_current_span().parent_span, None)
        self.assertEqual(get_current_span().root_span.trace_id, 't1')
        self.assertEqual(get_current_span().root_span.total_count, 1)
        self.assertTrue(get_current_span().is_root_sampling())
        s2 = Span('s2')
        s2.set_trace_id('t2')
        self.assertEqual(get_current_span().endpoint, 's2')
        self.assertEqual(get_current_span().trace_id, 't2')
        self.assertEqual(get_current_span().parent_span.trace_id, 't1')
        self.assertEqual(get_current_span().root_span.trace_id, 't1')
        self.assertEqual(get_current_span().root_span.total_count, 2)
        self.assertTrue(get_current_span().is_root_sampling())
        s3 = Span('s3')
        s3.set_trace_id('t3')
        self.assertEqual(get_current_span().endpoint, 's3')
        self.assertEqual(get_current_span().trace_id, 't3')
        self.assertEqual(get_current_span().parent_span.trace_id, 't2')
        self.assertEqual(get_current_span().root_span.trace_id, 't1')
        self.assertEqual(get_current_span().root_span.total_count, 3)
        self.assertTrue(get_current_span().is_root_sampling())
        s3.stop()
        self.assertEqual(get_current_span().endpoint, 's2')
        s2.set_trace_id('t1')
        s2.stop()
        self.assertEqual(s2.trace_id, 't1')
        s1.stop()
        self.assertIsNone(get_current_span())

    async def test_span_async(self):
        async def func1(ep):
            s2 = Span(ep)
            await asyncio.sleep(random.random() * 0.2)
            self.assertEqual(get_current_span().endpoint, ep)
            s2.stop()

        s1 = Span('r1')

        tasks = []
        for i in range(10):
            tasks.append(func1('s{i}'))

        await asyncio.gather(*tasks)

        self.assertEqual(get_current_span().root_span.total_count, 11)
        s1.stop()

    async def test_span_limit(self):
        for i in range(Span.MAX_NESTED_SPANS + 10):
            s = Span(f's{i}')
            if i < Span.MAX_NESTED_SPANS - 1:
                self.assertTrue(s.can_add_child())
                self.assertTrue(s.in_context)
            elif i == Span.MAX_NESTED_SPANS - 1:
                self.assertFalse(s.can_add_child())
                self.assertTrue(s.in_context)
            else:
                self.assertFalse(s.can_add_child())
                self.assertFalse(s.in_context)



import unittest
import logging
import sys
import time
from unittest.mock import patch, Mock

import graphsignal
from graphsignal.trace_sampler import TraceSampler

logger = logging.getLogger('graphsignal')


class TraceSamplerTest(unittest.TestCase):
    def setUp(self):
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            upload_on_shutdown=False,
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_group(self):
        limit = 2
        sampler = TraceSampler()
        for _ in range(limit):
            self.assertTrue(sampler.lock('g1', limit_per_interval=limit, limit_after=0))
            sampler.unlock()
        self.assertFalse(sampler.lock('g1', limit_per_interval=limit, limit_after=0))
        sampler.unlock()

    def test_limit_after(self):
        sampler = TraceSampler()
        for _ in range(10):
            self.assertTrue(sampler.lock('g1', limit_per_interval=0, limit_after=10))
            sampler.unlock()
        self.assertFalse(sampler.lock('g1', limit_per_interval=0, limit_after=10))
        sampler.unlock()

    def test_concurrent(self):
        sampler = TraceSampler()
        self.assertTrue(sampler.lock('g1'))
        self.assertFalse(sampler.lock('g1'))
        self.assertFalse(sampler.lock('g2'))
        sampler.unlock()
        self.assertTrue(sampler.lock('g2'))
        sampler.unlock()

    def test_reset(self):
        sampler = TraceSampler()

        self.assertTrue(sampler.lock('g1', limit_per_interval=1, limit_after=0))
        sampler.unlock()
        self.assertFalse(sampler.lock('g1', limit_per_interval=1, limit_after=0))
        sampler.unlock()

        sampler._last_reset_ts = time.time() - TraceSampler.MIN_INTERVAL_SEC - 10

        self.assertTrue(sampler.lock('g1', limit_per_interval=1, limit_after=0))
        sampler.unlock()
        self.assertFalse(sampler.lock('g1', limit_per_interval=1, limit_after=0))
        sampler.unlock()



import os
import sys
from setuptools import setup, find_packages, Extension

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

version = {}
with open('graphsignal/version.py') as fp:
    exec(fp.read(), version)

setup(
    name = 'graphsignal',
    version = version['__version__'],
    description = 'Graphsignal Tracer',
    long_description = read('README.md'),
    long_description_content_type = 'text/markdown',
    author = 'Graphsignal, Inc.',
    author_email = 'devops@graphsignal.com',
    url = 'https://graphsignal.com',
    license = 'Apache-2.0',
    keywords = [
        'inference monitoring',
        'model monitoring',
        'data monitoring',
        'exception tracking'
    ],
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development',
        'Topic :: System :: Monitoring',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence'
    ],
    python_requires='>=3.7',
    install_requires=[
        'protobuf>3.0'
    ],
    packages = find_packages(exclude=['*.sh', '*_test.py'])
)



import unittest
import logging
import sys
import time
from unittest.mock import patch, Mock

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.data.builtin_types import BuiltInTypesProfiler

logger = logging.getLogger('graphsignal')


class BuiltInTypesProfilerTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            upload_on_shutdown=False,
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_is_instance(self):
        profiler = BuiltInTypesProfiler()
        self.assertTrue(profiler.is_instance([]))

    def test_compute_stats(self):
        class TestObject:
            pass

        profiler = BuiltInTypesProfiler()
        stats = profiler.compute_stats([[0, -1, 1, 2.0, 3, float('nan'), float('inf'), TestObject()], [0, None, ''],[6,7]])
        self.assertEqual(
            stats.counts,
            {'byte_count': 260,
             'element_count': 13,
             'inf_count': 1,
             'nan_count': 1,
             'null_count': 1,
             'zero_count': 2,
             'empty_count': 1,
             'negative_count': 1,
             'positive_count': 6})
        self.assertEqual(stats.type_name, 'list')
        self.assertEqual(stats.shape, [3, 8])

    def test_compute_stats_none(self):
        profiler = BuiltInTypesProfiler()
        stats = profiler.compute_stats(None)
        self.assertEqual(
            stats.counts, 
            {'element_count': 1,
             'null_count': 1})
        self.assertEqual(stats.type_name, 'NoneType')

    def test_encode_sample(self):
        profiler = BuiltInTypesProfiler()
        preview = profiler.encode_sample(['text\n', 2.0, float('nan')])
        self.assertEqual(preview.content_type, 'application/json')
        self.assertEqual(preview.content_bytes, b'["text\\n", 2.0, NaN]')



import unittest
import logging
import sys
from unittest.mock import patch, Mock
import torch
from google.protobuf.json_format import MessageToJson
import pprint
import time
import random

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.traces import DEFAULT_OPTIONS
from graphsignal.recorders.process_recorder import ProcessRecorder

logger = logging.getLogger('graphsignal')

mem = []

class ProcessRecorderTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            upload_on_shutdown=False,
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_record(self):
        recorder = ProcessRecorder()
        recorder.setup()
        proto = signals_pb2.Trace()
        context = {}
        recorder.on_trace_start(proto, context, DEFAULT_OPTIONS)
        random.random()
        recorder.on_trace_stop(proto, context, DEFAULT_OPTIONS)
        recorder.on_trace_read(proto, context, DEFAULT_OPTIONS)

        ProcessRecorder.MIN_CPU_READ_INTERVAL_US = 0
        time.sleep(0.2)

        proto = signals_pb2.Trace()
        context = {}
        recorder.on_trace_start(proto, context, DEFAULT_OPTIONS)
        for _ in range(100000):
            random.random()
        global mem
        mem = [1]*100000
        recorder.on_trace_stop(proto, context, DEFAULT_OPTIONS)
        recorder.on_trace_read(proto, context, DEFAULT_OPTIONS)

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(proto))

        self.assertNotEqual(proto.node_usage.hostname, '')
        self.assertNotEqual(proto.node_usage.ip_address, '')
        self.assertNotEqual(proto.process_usage.pid, '')
        if sys.platform != 'win32':
            self.assertTrue(proto.node_usage.mem_total > 0)
            self.assertTrue(proto.node_usage.mem_used > 0)
            self.assertTrue(proto.process_usage.cpu_usage_percent > 0)
            self.assertTrue(proto.process_usage.current_rss > 0)
            self.assertTrue(proto.process_usage.max_rss > 0)
            self.assertTrue(proto.process_usage.vm_size > 0)

        recorder.on_metric_update()

        store = graphsignal._agent.metric_store()
        metric_tags =  {'deployment': 'd1', 'hostname': proto.node_usage.hostname}
        key = store.metric_key('system', 'process_cpu_usage', metric_tags)
        self.assertTrue(store.metrics[key].gauge > 0)
        key = store.metric_key('system', 'process_memory', metric_tags)
        self.assertTrue(store.metrics[key].gauge > 0)
        key = store.metric_key('system', 'virtual_memory', metric_tags)
        self.assertTrue(store.metrics[key].gauge > 0)
        key = store.metric_key('system', 'node_memory_used', metric_tags)
        self.assertTrue(store.metrics[key].gauge > 0)



import sys
import runpy
import graphsignal


def main():
    # remove own path
    sys.argv.pop(0)

    # detect if we are running as a module
    is_module = False
    if len(sys.argv) >= 1:
        if sys.argv[0] == '-m':
            is_module = True
            sys.argv.pop(0)
    else:
        sys.exit(1)

    if len(sys.argv) < 1:
        sys.exit(1)

    # initialize graphsignal   
    graphsignal.configure()

    # run it as normal
    module_or_script = sys.argv[0]
    if is_module:
        runpy.run_module(module_or_script, run_name='__main__')
    else:
        exec(open(module_or_script).read(), globals(), locals())


if __name__ == "__main__":
    main()



from typing import Dict, Any, Union, Optional
import os
import logging
import atexit
import functools
import asyncio

from graphsignal.version import __version__
from graphsignal.agent import Agent
from graphsignal.traces import Trace, TraceOptions

logger = logging.getLogger('graphsignal')

_agent = None


def _check_configured():
    global _agent
    if not _agent:
        raise ValueError(
            'Tracer not configured, call graphsignal.configure() first')


def _check_and_set_arg(
        name, value, is_str=False, is_int=False, is_bool=False, is_kv=False, required=False, max_len=None):
    env_name = 'GRAPHSIGNAL_{0}'.format(name.upper())

    if not value and env_name in os.environ:
        value = os.environ[env_name]
        if value:
            if is_str:
                if max_len and len(value) > max_len:
                    raise ValueError('configure: invalid format, expected string with max length {0}: {1}'.format(max_len, name))
            if is_int:
                try:
                    value = int(value)
                except:
                    raise ValueError('configure: invalid format, expected integer: {0}'.format(name))
            elif is_bool:
                value = bool(value)
            elif is_kv:
                try:
                    value = dict([el.strip(' ') for el in kv.split('=')] for kv in value.split(','))
                except:
                    raise ValueError('configure: invalid format, expected comma-separated key-value list (k1=v1,k2=v2): {0}'.format(name))

    if not value and required:
        raise ValueError('configure: missing argument: {0}'.format(name))

    return value


def configure(
        api_key: Optional[str] = None,
        api_url: Optional[str] = None,
        deployment: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
        auto_instrument: Optional[bool] = True,
        record_data_samples: Optional[bool] = True,
        upload_on_shutdown: Optional[bool] = True,
        debug_mode: Optional[bool] = False) -> None:
    global _agent

    if _agent:
        logger.warning('Tracer already configured')
        return

    debug_mode = _check_and_set_arg('debug_mode', debug_mode, is_bool=True)
    if debug_mode:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.WARNING)
    api_key = _check_and_set_arg('api_key', api_key, is_str=True, required=True)
    api_url = _check_and_set_arg('api_url', api_url, is_str=True, required=False)
    deployment = _check_and_set_arg('deployment', deployment, is_str=True, required=True)
    tags = _check_and_set_arg('tags', tags, is_kv=True, required=False)

    _agent = Agent(
        api_key=api_key,
        api_url=api_url,
        deployment=deployment,
        tags=tags,
        auto_instrument=auto_instrument,
        record_data_samples=record_data_samples,
        upload_on_shutdown=upload_on_shutdown,
        debug_mode=debug_mode)
    _agent.setup()

    atexit.register(shutdown)

    logger.debug('Tracer configured')


def set_tag(key: str, value: str) -> None:
    _check_configured()

    if not key:
        logger.error('set_tag: key must be provided')
        return
    if value is None:
        logger.error('set_tag: value must be provided')
        return

    if _agent.tags is None:
        _agent.tags = {}

    if len(_agent.tags) > Trace.MAX_RUN_TAGS:
        logger.error('set_tag: too many tags (>{0})'.format(Trace.MAX_RUN_TAGS))
        return

    _agent.tags[key] = value


def set_context_tag(key: str, value: str) -> None:
    _check_configured()

    if not key:
        logger.error('set_context_tag: key must be provided')
        return
    if value is None:
        logger.error('set_context_tag: value must be provided')
        return

    tags = _agent.context_tags.get()
    if len(tags) > Trace.MAX_RUN_TAGS:
        logger.error('set_context_tag: too many tags (>{0})'.format(Trace.MAX_TRACE_TAGS))
        return

    tags[key] = value
    _agent.context_tags.set(tags)


def start_trace(
        endpoint: str,
        tags: Optional[Dict[str, str]] = None,
        options: Optional[TraceOptions] = None) -> Trace:
    _check_configured()

    return Trace(endpoint=endpoint, tags=tags, options=options)


def trace_function(
        func=None, *,
        endpoint: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None,
        options: Optional[TraceOptions] = None):
    if func is None:
        return functools.partial(trace_function, endpoint=endpoint, tags=tags, options=options)

    if endpoint is None:
        endpoint_or_name = func.__name__
    else:
        endpoint_or_name = endpoint

    if asyncio.iscoroutinefunction(func):
        @functools.wraps(func)
        async def tf_async_wrapper(*args, **kwargs):
            with start_trace(endpoint=endpoint_or_name, tags=tags, options=options):
                return await func(*args, **kwargs)
        return tf_async_wrapper
    else:
        @functools.wraps(func)
        def tf_wrapper(*args, **kwargs):
            with start_trace(endpoint=endpoint_or_name, tags=tags, options=options):
                return func(*args, **kwargs)
        return tf_wrapper


def upload(block=False) -> None:
    _check_configured()

    _agent.upload(block=block)


def shutdown() -> None:
    global _agent
    if not _agent:
        return

    atexit.unregister(shutdown)
    _agent.shutdown()
    _agent = None

    logger.debug('Tracer shutdown')


__all__ = [
    '__version__',
    'configure',
    'upload',
    'shutdown',
    'start_trace',
    'function_trace',
    'Trace',
    'TraceOptions'
]


import logging

logger = logging.getLogger('graphsignal')


class MissingValueDetector:
    MISSING_COUNTERS = {
        'null_count',
        'nan_count',
        'inf_count',
        'empty_count'
    }

    def __init__(self):
        pass

    def detect(self, data_name, counts):
        for counter_name in self.MISSING_COUNTERS:
            if counter_name in counts:
                if counts[counter_name] > 0:
                    return True

        return False


import logging
import sys
import json

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.data.base_data_profiler import BaseDataProfiler, DataStats, DataSample

logger = logging.getLogger('graphsignal')


class BuiltInTypesProfiler(BaseDataProfiler):
    def __init__(self):
        super().__init__()

    def is_instance(self, data):
        return isinstance(data, (list, dict, tuple, set, str, int, float, complex)) or data is None

    def compute_stats(self, data):
        counts = {}
        counts['byte_count'] = _size(data)
        counts['element_count'] = _count_true(data, lambda elem: True)
        counts['null_count'] = _count_true(data, lambda elem: elem is None)
        counts['nan_count'] = _count_true(data, lambda elem: elem != elem)
        counts['inf_count'] = _count_true(data, lambda elem: elem == float("inf") or elem == float("-inf"))
        counts['zero_count'] = _count_true(data, lambda elem: elem == 0)
        counts['empty_count'] = _count_true(data, lambda elem: elem == '')
        counts['negative_count'] = _count_true(data, lambda elem: isinstance(elem, (int, float)) and elem < 0)
        counts['positive_count'] = _count_true(data, lambda elem: isinstance(elem, (int, float)) and elem > 0)
        return DataStats(type_name=type(data).__name__, shape=_shape(data), counts=counts)

    def encode_sample(self, data):
        return DataSample(content_type='application/json', content_bytes=json.dumps(data).encode('utf-8'))


def _shape(data):
    if isinstance(data, list):
        if len(data) == 0:
            return [0]
        else:
            return [len(data)] + _shape(data[0])
    return []


def _size(data):
    if isinstance(data, dict):
        return sum([_size(elem) for elem in data.values()])
    elif isinstance(data, (list, tuple, set)):
        return sum([_size(elem) for elem in data])
    elif isinstance(data, (str, bytes)):
        return len(data)
    elif isinstance(data, (int, float, complex)):
        return sys.getsizeof(data)
    return 0


def _count_true(data, filter_func):
    if isinstance(data, dict):
        return sum([_count_true(elem, filter_func) for elem in data.values()])
    elif isinstance(data, (list, tuple, set)):
        return sum([_count_true(elem, filter_func) for elem in data])
    else:
        return sum([filter_func(data)])
